/* gIDE indent plugin
 * 
 * Run indent on the current document.
 * 
 * Copyright (C) 2001 Andre van Dijk "Unseen" <andre@unseen.demon.nl>
 * Copyright (C) 2001 Roberto Majadas "telemaco" <phoenix@nova.es>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.  
 */

#include <config.h>

#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <wait.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <liboaf/liboaf.h>
#include <libgide/libgide.h>
#include <gdl/GDL.h>

#define DEBUG
#define SAMPLE_COMPONENT_IID "OAFIID:GNOME_Development_Plugin:indent"
#define INDENT_SH_NAME "sh"
#define INDENT_SH_PATH "/bin/sh"

#define STYLE_CUSTOM 0
#define STYLE_GNU 1
#define STYLE_KR 2
#define STYLE_BERKELEY 3

typedef struct _indent_dialog indent_dialog;
typedef struct _indent_info indent_info;

struct _indent_dialog
{
  GtkWidget *dialog;
  GtkWidget *custom_radio;
};

struct _indent_info
{
  gint indent_level;
  gint coding_style;
  gchar *temp_file_name;
  GideTool *tool;
};

static indent_info *indentinfo;
static indent_dialog *indentdialog;

/*
 * Get current document as a GString
 */
GString *indent_getdocument (GideTool * tool);

/*
 * Execute indent on current document
 */
static void indent_execute ();

/*
 * Delete current document text and load temp file instead
 */
static void indent_load_temp_file ();

/*
 * Create dialog and show
 */
static void indent_create_dialog ();

/*
 * Close dialog and clean up
 */
static void indent_close_dialog (GtkWidget * widget, gpointer data);

/*
 * Create the notebook for the dialog
 */
static void indent_create_notebook ();

/* Dialog Callbacks */
void on_coding_gnu_toggled (GtkToggleButton * togglebutton,
			    gpointer user_data);
void on_coding_kr_toggled (GtkToggleButton * togglebutton,
			   gpointer user_data);
void on_coding_berkeley_toggled (GtkToggleButton * togglebutton,
				 gpointer user_data);
void on_coding_custom_toggled (GtkToggleButton * togglebutton,
			       gpointer user_data);
void on_indent_level_changed (GtkEditable * editable, gpointer user_data);

void
on_coding_gnu_toggled (GtkToggleButton * togglebutton, gpointer user_data)
{
#ifdef DEBUG
  g_print ("GNU coding style selected\n");
#endif
  indentinfo->coding_style = STYLE_GNU;
}


void
on_coding_kr_toggled (GtkToggleButton * togglebutton, gpointer user_data)
{
#ifdef DEBUG
  g_print ("K&R coding style selected\n");
#endif
  indentinfo->coding_style = STYLE_KR;

}


void
on_coding_berkeley_toggled (GtkToggleButton * togglebutton,
			    gpointer user_data)
{
#ifdef DEBUG
  g_print ("Berkeley coding style selected\n");
#endif
  indentinfo->coding_style = STYLE_BERKELEY;

}

void
on_coding_custom_toggled (GtkToggleButton * togglebutton, gpointer user_data)
{
#ifdef DEBUG
  g_print ("Custom coding style selected\n");
#endif
  indentinfo->coding_style = STYLE_CUSTOM;
}

void
on_indent_level_changed (GtkEditable * editable, gpointer user_data)
{
  /* Toggle custom coding style when indent level is changed */
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON
				(indentdialog->custom_radio), TRUE);
  /* Set indent_level parameter from spinner widget */
  indentinfo->indent_level = atoi (gtk_editable_get_chars (editable, 0, -1));
#ifdef DEBUG
  g_print ("Indent_level: %d\n", indentinfo->indent_level);
#endif
}

static void
indent_close_dialog (GtkWidget * widget, gpointer data)
{
  gnome_dialog_close (GNOME_DIALOG (indentdialog->dialog));
  g_free (indentinfo->temp_file_name);
  g_free (indentinfo);
  g_free (indentdialog);
}

static void
indent_load_temp_file ()
{
  gint fd, i;
  gint pos = 0;
  gchar buf[1024];

  if ((fd = open (indentinfo->temp_file_name, O_RDONLY)) == -1)
    {
      gnome_error_dialog ("Error opening temporary file.");
      return;
    }

  /* Delete current document text */
  gide_delete_text (indentinfo->tool, 0,
		    gide_get_document_length (indentinfo->tool));

  /* Read temporary file into buffer */
  while ((i = read (fd, buf, 1023)) > 0)
    {
      buf[i] = '\0';
      gide_insert_text_at_pos (indentinfo->tool, pos, buf);
      pos += i;
    }
}

static void
indent_execute ()
{
  gint leftpipe[2];
  gint fd;
  gint pid;
  gchar *arg[4];
  gchar *command_to_run;
  gchar *coding_style_arg;
  GString *buffer_to_pipe;

#ifdef DEBUG
  g_print ("indent_execute\n");
#endif

  /* Create pipe */
  if (pipe (leftpipe) == -1)
    {
      gnome_error_dialog
	("Unable to create communication channel for indent.");
      indent_close_dialog (NULL, NULL);
      return;
    }

  /* Get current document text */
  buffer_to_pipe = indent_getdocument (indentinfo->tool);
  if (buffer_to_pipe == NULL)
    {
      /* No document so not much to do */
      indent_close_dialog (NULL, NULL);
      return;
    }

  /* Construct command string with options */
  switch (indentinfo->coding_style)
    {
    case STYLE_GNU:
      coding_style_arg = "-gnu";
      break;
    case STYLE_KR:
      coding_style_arg = "-kr";
      break;
    case STYLE_BERKELEY:
      coding_style_arg = "-orig";
      break;
    default:
      /* Custom style */
      if (indentinfo->indent_level > 0)
	coding_style_arg =
	  g_strdup_printf ("-i %d", indentinfo->indent_level);
      else
	coding_style_arg = "";
      break;
    }
  command_to_run = g_strdup_printf (" indent -st %s ", coding_style_arg);
#ifdef DEBUG
  g_print ("indent_execute: command to run: %s\n", command_to_run);
#endif

  pid = fork ();
  if (pid == 0)
    {				/* Child process */

      /* Stdin reads from left pipe */
      dup2 (leftpipe[0], 0);
      close (leftpipe[0]);
      close (leftpipe[1]);

      /* Stdout writes to temp file */
      fd = open (indentinfo->temp_file_name, O_RDWR | O_CREAT, S_IRWXU);
      if (fd == -1)
	{
	  g_warning ("Unable to create temporary file\n");
	  return;
	}
      dup2 (fd, 1);
      close (fd);

      /* Run shell with indent */
      arg[0] = INDENT_SH_NAME;
      arg[1] = "-c";
      arg[2] = command_to_run;
      arg[3] = NULL;

      execvp (INDENT_SH_PATH, arg);

      g_warning ("An undetermined PIPE problem occurred\n");
      return;
    }
  else if (pid == -1)
    {
      gnome_error_dialog ("Problem running indent.");
    }
  else
    {
      close (leftpipe[0]);
#ifdef DEBUG
      g_print ("Writing %d bytes\n", buffer_to_pipe->len);
#endif

      /* Write document to pipe */
      if (buffer_to_pipe->len > 0)
	{
	  fcntl (leftpipe[1], F_SETFD, FD_CLOEXEC);
	  write (leftpipe[1], buffer_to_pipe->str, buffer_to_pipe->len);
	  close (leftpipe[1]);
	}
      waitpid (pid, NULL, 0);	/* Maybe should process ident return value */
      indent_load_temp_file ();
    }
  g_string_free (buffer_to_pipe, TRUE);
  indent_close_dialog (NULL, NULL);
}

static void
indent_create_notebook ()
{
  GtkWidget *notebook1;
  GtkWidget *vbox1;
  GSList *coding_group = NULL;
  GtkWidget *coding_gnu;
  GtkWidget *coding_kr;
  GtkWidget *coding_berkeley;
  GtkWidget *radiobutton1;
  GtkWidget *coding;
  GtkWidget *vbox2;
  GtkWidget *hbox2;
  GtkWidget *label4;
  GtkObject *indent_level_adj;
  GtkWidget *indent_level;
  GtkWidget *custom;
  GtkWidget *dialog_vbox;

  dialog_vbox = GNOME_DIALOG (indentdialog->dialog)->vbox;
  gtk_widget_show (dialog_vbox);

  notebook1 = gtk_notebook_new ();
  gtk_widget_show (notebook1);
  gtk_container_add (GTK_CONTAINER (dialog_vbox), notebook1);

  vbox1 = gtk_vbox_new (FALSE, 0);
  gtk_widget_ref (vbox1);
  gtk_object_set_data_full (GTK_OBJECT (indentdialog->dialog), "vbox1", vbox1,
			    (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (vbox1);
  gtk_container_add (GTK_CONTAINER (notebook1), vbox1);

  coding_gnu = gtk_radio_button_new_with_label (coding_group, _("GNU"));
  coding_group = gtk_radio_button_group (GTK_RADIO_BUTTON (coding_gnu));
  gtk_widget_ref (coding_gnu);
  gtk_object_set_data_full (GTK_OBJECT (indentdialog->dialog), "coding_gnu",
			    coding_gnu, (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (coding_gnu);
  gtk_box_pack_start (GTK_BOX (vbox1), coding_gnu, FALSE, FALSE, 0);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (coding_gnu), TRUE);

  coding_kr =
    gtk_radio_button_new_with_label (coding_group, _("Kernighan & Ritchie"));
  coding_group = gtk_radio_button_group (GTK_RADIO_BUTTON (coding_kr));
  gtk_widget_ref (coding_kr);
  gtk_object_set_data_full (GTK_OBJECT (indentdialog->dialog), "coding_kr",
			    coding_kr, (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (coding_kr);
  gtk_box_pack_start (GTK_BOX (vbox1), coding_kr, FALSE, FALSE, 0);

  coding_berkeley =
    gtk_radio_button_new_with_label (coding_group, _("Original Berkeley"));
  coding_group = gtk_radio_button_group (GTK_RADIO_BUTTON (coding_berkeley));
  gtk_widget_ref (coding_berkeley);
  gtk_object_set_data_full (GTK_OBJECT (indentdialog->dialog),
			    "coding_berkeley", coding_berkeley,
			    (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (coding_berkeley);
  gtk_box_pack_start (GTK_BOX (vbox1), coding_berkeley, FALSE, FALSE, 0);

  radiobutton1 = gtk_radio_button_new_with_label (coding_group, _("Custom"));
  coding_group = gtk_radio_button_group (GTK_RADIO_BUTTON (radiobutton1));
  gtk_widget_ref (radiobutton1);
  gtk_object_set_data_full (GTK_OBJECT (indentdialog->dialog), "radiobutton1",
			    radiobutton1,
			    (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (radiobutton1);
  indentdialog->custom_radio = radiobutton1;
  gtk_box_pack_start (GTK_BOX (vbox1), radiobutton1, FALSE, FALSE, 0);

  coding = gtk_label_new (_("Coding Style"));
  gtk_widget_ref (coding);
  gtk_object_set_data_full (GTK_OBJECT (indentdialog->dialog), "coding",
			    coding, (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (coding);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook1),
			      gtk_notebook_get_nth_page (GTK_NOTEBOOK
							 (notebook1), 0),
			      coding);

  vbox2 = gtk_vbox_new (FALSE, 0);
  gtk_widget_ref (vbox2);
  gtk_object_set_data_full (GTK_OBJECT (indentdialog->dialog), "vbox2", vbox2,
			    (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (vbox2);
  gtk_container_add (GTK_CONTAINER (notebook1), vbox2);

  hbox2 = gtk_hbox_new (FALSE, 0);
  gtk_widget_ref (hbox2);
  gtk_object_set_data_full (GTK_OBJECT (indentdialog->dialog), "hbox2", hbox2,
			    (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (hbox2);
  gtk_box_pack_start (GTK_BOX (vbox2), hbox2, TRUE, TRUE, 0);

  label4 = gtk_label_new (_("Level:"));
  gtk_widget_ref (label4);
  gtk_object_set_data_full (GTK_OBJECT (indentdialog->dialog), "label4",
			    label4, (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label4);
  gtk_box_pack_start (GTK_BOX (hbox2), label4, FALSE, FALSE, 0);

  indent_level_adj = gtk_adjustment_new (1, 0, 100, 1, 10, 10);
  indent_level =
    gtk_spin_button_new (GTK_ADJUSTMENT (indent_level_adj), 1, 0);
  gtk_widget_ref (indent_level);
  gtk_object_set_data_full (GTK_OBJECT (indentdialog->dialog), "indent_level",
			    indent_level,
			    (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (indent_level);
  gtk_box_pack_start (GTK_BOX (hbox2), indent_level, FALSE, FALSE, 0);
  gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (indent_level), TRUE);

  custom = gtk_label_new (_("Custom\t"));
  gtk_widget_ref (custom);
  gtk_object_set_data_full (GTK_OBJECT (indentdialog->dialog), "custom",
			    custom, (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (custom);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook1),
			      gtk_notebook_get_nth_page (GTK_NOTEBOOK
							 (notebook1), 1),
			      custom);

  gtk_signal_connect (GTK_OBJECT (coding_gnu), "toggled",
		      GTK_SIGNAL_FUNC (on_coding_gnu_toggled), NULL);
  gtk_signal_connect (GTK_OBJECT (coding_kr), "toggled",
		      GTK_SIGNAL_FUNC (on_coding_kr_toggled), NULL);
  gtk_signal_connect (GTK_OBJECT (coding_berkeley), "toggled",
		      GTK_SIGNAL_FUNC (on_coding_berkeley_toggled), NULL);
  gtk_signal_connect (GTK_OBJECT (radiobutton1), "toggled",
		      GTK_SIGNAL_FUNC (on_coding_custom_toggled), NULL);
  gtk_signal_connect_after (GTK_OBJECT (indent_level), "changed",
			    GTK_SIGNAL_FUNC (on_indent_level_changed), NULL);

}

static void
indent_create_dialog ()
{
  /* Create a OK or Cancel Dialog */
  indentdialog->dialog =
    gnome_dialog_new (_("Indent Options"), GNOME_STOCK_BUTTON_OK,
		      GNOME_STOCK_BUTTON_CANCEL, NULL);
  gtk_window_set_policy (GTK_WINDOW (indentdialog->dialog), FALSE, FALSE,
			 FALSE);

  gtk_widget_show (GNOME_DIALOG (indentdialog->dialog)->vbox);
  /* Add options notebook */
  indent_create_notebook ();
  /* OK button runs indent */
  gnome_dialog_button_connect (GNOME_DIALOG (indentdialog->dialog), 0,
			       GTK_SIGNAL_FUNC (indent_execute), NULL);
  /* Cancel closes dialog */
  gnome_dialog_button_connect (GNOME_DIALOG (indentdialog->dialog), 1,
			       GTK_SIGNAL_FUNC (indent_close_dialog), NULL);
  gtk_signal_connect (GTK_OBJECT (indentdialog->dialog), "delete_event",
		      GTK_SIGNAL_FUNC (indent_close_dialog), NULL);
  /* Show Dialog */
  gtk_widget_show (indentdialog->dialog);
}


/* 
 * Return text of current document 
 */
GString *
indent_getdocument (GideTool * tool)
{
  GString *ret = NULL;

  if (gide_get_document_length (tool) > 0)
    ret =
      g_string_new (gide_get_document_chars
		    (tool, 0, gide_get_document_length (tool)));
  return ret;
}

static void
indent (GtkWidget * widget, gpointer data)
{
  /* Create indentinfo */
  indentinfo = g_new (indent_info, 1);
  /* Create indentdialog */
  indentdialog = g_new (indent_dialog, 1);
  /* Create temp filename */
  indentinfo->temp_file_name = tempnam (g_get_tmp_dir (), "gide-");
  /* Init indent level */
  indentinfo->indent_level = 0;
  /* Set default coding style */
  indentinfo->coding_style = STYLE_GNU;
  /* Safe reference to tool */
  indentinfo->tool = GIDE_TOOL (data);
  /* Create and show dialog */
  indent_create_dialog ();
}

/*
 * Define the verbs in this plugin
 */
static BonoboUIVerb verbs[] = {
  BONOBO_UI_UNSAFE_VERB ("Indent", indent),
  BONOBO_UI_VERB_END
};

/*
 * Boilerplate initialization function
 */
static gboolean
impl_init (GideTool * tool, gpointer closure)
{
  BonoboUIComponent *uic;
  CORBA_Environment ev;

  g_return_val_if_fail (tool != NULL, FALSE);
  g_return_val_if_fail (GIDE_IS_TOOL (tool), FALSE);

  CORBA_exception_init (&ev);

  uic = bonobo_ui_component_new ("gide-indent-plugin");
  bonobo_ui_component_set_container (uic, tool->ui_container);
  bonobo_ui_util_set_ui (uic, GIDE_DATADIR,
			 "gide-indent-plugin.xml", "gide-indent-plugin");

  bonobo_ui_component_add_verb_list_with_data (uic, verbs, tool);

  CORBA_exception_free (&ev);

  return TRUE;
}

/*
 * Boilerplace cleanup function
 */
static void
impl_cleanup (GideTool * tool, gpointer closure)
{
}

/*
 * Boilerplate creation function
 */
static CORBA_Object
make_object (PortableServer_POA poa, const char *iid, gpointer impl_ptr,
	     CORBA_Environment * ev)
{
  GideTool *tool;

  tool = gide_tool_new (impl_init, impl_cleanup, NULL);

  oaf_plugin_use (poa, impl_ptr);

  return BONOBO_OBJREF (tool->props);
}

/*
 * List of plugins
 */
static const OAFPluginObject plugin_list[] = {
  {
   SAMPLE_COMPONENT_IID,
   make_object},
  {
   NULL}
};

/*
 * Plugin description
 */
const OAFPlugin OAF_Plugin_info = {
  plugin_list,
  "gIDE Indent Plugin"
};
